// Project: Delta.ContentSystem

//#define LOG_STUFF

using System; 
using System.Collections.Generic;
using System.IO;
using Delta.ContentSystem.Xml;
using Delta.Engine;
using Delta.Engine.Dynamic;
using Delta.Utilities.Datatypes;
using Delta.Utilities.Datatypes.Graphics;
using Delta.Utilities.Helpers;
#if ANDROID
using Delta.Platforms.MonoDroid;
#endif

namespace Delta.ContentSystem
{
	/// <summary>
	/// Base content manager class, different implementations might work
	/// completely different (just local files, downloading files from the web
	/// or services, etc.), but in all cases we need local files to load the
	/// actual content from disk (where ever it came from), for details see the
	/// GetLocalFilePath method. Delayed loading is also possible via events:
	/// Content first uses the old content or dummy content if we start with
	/// nothing, then once we notify the Content again and load the real content
	/// we have downloaded from a service for example in the meantime.
	/// </summary>
	public abstract class ContentManager : DynamicModule, IDisposable
	{
		#region Constants
		/// <summary>
		/// Same constant as in Localization.English, but we have no access to
		/// that class here.
		/// </summary>
		protected const string English = "English";

		/// <summary>
		/// Helper to ignore all warnings in LoadContentList for the ContentServer!
		/// </summary>
		public static bool IgnoreNoContentForContentServer = false;
		#endregion

		#region Static (ServerRawDirectory, PreloadAll and KillUnusedContent)
		#region ServerRawDirectory
		/// <summary>
		/// Server raw directory, can be changed in the ContentGenerator!
		/// </summary>
		public static string ServerRawDirectory =
			@"\\blub\DummyRawContent";
		#endregion

		#region ServerPlatformDirectory
		/// <summary>
		/// Server platform directory, can be changed in the ContentGenerator too.
		/// </summary>
		public static string ServerPlatformDirectory =
			@"\\blub\DummyPlatformContent";
		#endregion

		#region ServerProjectName
		/// <summary>
		/// Server project name, used for generating content and excluding
		/// other projects in the ContentGenerator.
		/// </summary>
		public static string ServerProjectName =
			@"Engine";
		#endregion

		#region PreloadAll
		/// <summary>
		/// Helper method to reload all
		/// </summary>
		public static void PreloadAll()
		{
			// Check all content if we have anything not fully loaded yet.
			// Note: This is obviously risky because if the content list is huge
			// we will load all this content, the streaming in Run must make sure
			// that old unused content was already removed!
			foreach (Content content in Instance.usedContentClasses)
			{
				// Use the CachedData property (which is slow btw) to check if the
				// content already has been loaded. If not, then force to load.
				if (content.loadInRun ||
					content.CachedData == null)
				{
					content.ForceLoad();
				}
			} // foreach
		} // PreloadAll()
		#endregion

		#region KillUnusedContent
		/// <summary>
		/// Kill unused content
		/// </summary>
		public static void KillUnusedContent()
		{
			Instance.TryToKillUnusedContent();
		} // KillUnusedContent()
		#endregion
		#endregion

		#region Public
		#region Instance
		/// <summary>
		/// Content manager instance we are going to use. See
		/// Delta.ContentSystem.Client for the only current implementation.
		/// </summary>
		public static ContentManager contentManagerInstance = null;
		/// <summary>
		/// Content manager instance, required to load all content files. Behind
		/// the scenes this is actually a very complex process, especially on
		/// Windows where the Content.Data file can be re-generated if we don't
		/// find a content file (via the ContentServer that processes RawContent
		/// and builds platform specific content). This instance is currently
		/// only needed from the Content constructor to find each ContentData.
		/// </summary>
		public static ContentManager Instance
		{
			get
			{
				if (contentManagerInstance == null)
				{
					contentManagerInstance = Factory.Create<ContentManager>();
					if (contentManagerInstance == null)
					{
						// Without content no real application can start, we need this!
						Application.Error("The application will quit now because no " +
							"ContentManager module is available, it is impossible to " +
							"start the content manager system. Please provide a content " +
							"manager module (default one is Delta.ContentSystem.Client).");
					}
#if ANDROID
					MonoDroidGameView.InvalidateContent = delegate()
					{
						Instance.Dispose();
					};
#endif
				} // if
				return contentManagerInstance;
			} // get
		} // Instance
		#endregion

		#region NumberOfContentFiles
		/// <summary>
		/// Returns the number of content files that have been crawled and can
		/// be used. For the current number of loaded content files use the
		/// LoadedContentFiles property.
		/// </summary>
		public int NumberOfContentFiles
		{
			get
			{
				// Count total content files with all the dictionary lists
				int totalContentFiles = 0;
				foreach (List<ContentData> datas in contentDataDictionary.Values)
				{
					totalContentFiles += datas.Count;
				}
				return totalContentFiles;
			}
		} // NumberOfContentFiles
		#endregion

		//TODO: clean up here!
		#region LoadedContentFiles
		/// <summary>
		/// Number of currently loaded content files (all ContentData objects that
		/// have the IsLoaded property set to true).
		/// </summary>
		public int LoadedContentFiles
		{
			get
			{
				// Just count all the loaded content data lists 
				int numberOfLoadedObjects = 0;
				//obs: foreach (List<ContentData> datas in contentDataDictionary.Values)
				foreach (NativeData datas in alreadyLoadedNativeData.Values)
				{
					foreach (Content data in datas.usedContents)
					{
						//always true: if (data.IsLoaded)
						//{
						numberOfLoadedObjects++;
						//}
					}
				}
				return numberOfLoadedObjects;
			} // get
		} // LoadedContentFiles
		#endregion

		#region TotalLoadedContentSize
		/// <summary>
		/// Total loaded content size in bytes, will go through all the loaded
		/// ContentData objects and count up the file size. Used for the Profiler.
		/// </summary>
		public int TotalLoadedContentSize
		{
			get
			{
				// Just count to total loaded content data size
				int totalBytes = 0;
				//obs: foreach (List<ContentData> datas in contentDataDictionary.Values)
				foreach (NativeData datas in alreadyLoadedNativeData.Values)
				{
					//we only care about the size once, not for each atlas image:
					//foreach (Content data in datas.usedContents)
					//{
					//always true: if (data.IsLoaded)
					//{
					if (datas.usedContents.Count > 0)
					{
						totalBytes += datas.usedContents[0].data.FileSize;
					}
					//}
				}
				return totalBytes;
			} // get
		} // TotalLoadedContentSize
		#endregion

		#region TotalLoadedContentInMb
		/// <summary>
		/// Total loaded content size in megabyte
		/// </summary>
		public static int TotalLoadedContentInMb
		{
			get
			{
				return Instance.TotalLoadedContentSize / (1024 * 1024);
			}
		} // TotalLoadedContentInMb
		#endregion
		#endregion

		#region Private
		/// <summary>
		/// Internal list of content data objects that will be loaded once this
		/// content manager is created (in the derived implementation classes).
		/// Instead of using a flat list (which works fine for a few content files)
		/// we are using a dictionary, which we can search case insensitive and
		/// this way searches are a lot faster for many files (more than 10x).
		/// It is important to note that this ContentList only contains whatever
		/// local content files are available. Content names that are not found
		/// will result in requesting them from the ContentServer if allowed on
		/// this platform or just in empty ContentData objects that will just use
		/// their Fallback implementation (see Log for warnings about that).
		/// </summary>
		protected Dictionary<string, List<ContentData>> contentDataDictionary =
			new Dictionary<string, List<ContentData>>(//obs:nameComparer);
			StringComparer.InvariantCultureIgnoreCase);
		/// <summary>
		/// Since we only cache each ContentData object and keep a list of used
		/// content classes, we can only cache each content request and update all
		/// the loaded content classes (e.g. for language changes). But we cannot
		/// return the same content data object for an already loaded atlas
		/// texture because the ContentData object will be a different one for
		/// each of the used images in the atlas, even when they share all the
		/// same underlying native object (the loaded native atlas texture).
		/// The key for this dictionary is the LocalFilePath from ContentData and
		/// the value is the actual loaded object (or null if not loaded yet).
		/// This dictionary is mainly used from ContentData, but having access
		/// to it here makes our caching life easier (e.g. removing loaded data
		/// and disposing it in case of a language switch or content update).
		/// </summary>
		internal Dictionary<string, NativeData> alreadyLoadedNativeData =
			new Dictionary<string, NativeData>();//obs:(nameComparer);
		//TODO: we should use disposing and streaming, etc. but that is all disabled right now!
		//internal Dictionary<string, NativeData> alreadyLoadedNativeData =
		//	new Dictionary<string, NativeData>();
		/// <summary>
		/// Remember all content classes that called GetContentData, which might
		/// need to be updated in case of a language change or a content update.
		/// Each content class will be removed again from this list in the
		/// RemoveContentClass method. This list can grow huge, but it is not
		/// often used, only if content updates or language switches happen!
		/// </summary>
		protected List<Content> usedContentClasses = new List<Content>();
		/// <summary>
		/// Keep a static reference of each fallback object for each content type.
		/// This way each of the content classes does not need to implement this
		/// cache object himself for fallback content. This is only used when some
		/// content loading failed (see log), if everything is working fine, this
		/// dictionary is never used and LoadFallbackData is never called. Used
		/// only in the Content.Load method, but we need to clear it here too.
		/// </summary>
		internal Dictionary<ContentType, object> cachedFallbackData =
			new Dictionary<ContentType, object>();//obs:(typeComparer);

		/// <summary>
		/// Name of the current project used for the PlatformConfiguration,
		/// which is send to the ContentServer.
		/// </summary>
		protected string projectName;
		#endregion

		#region Constructor
		/// <summary>
		/// Create content manager, will just attach to the language change event
		/// and derived classes will do the same for content update events.
		/// </summary>
		protected ContentManager()
			// Use Application as our base, it does not really matter when this code
			// is executed, Run is unused, but might be useful for streaming later
			: base("ContentManager", typeof(Application))
		{
			// Cache the project name, this will also trigger setting the
			// Log.ProgramName correctly, which is good for all early content
			// errors we might get from here on.
			projectName = Application.ProjectName;

			Settings.LanguageChanged += OnLanguageChanged;
#if WINDOWS
			// This is only supported on Windows, all other platforms can either not
			// create new directories or it makes no sense anyway because we cannot
			// download new content dynamically!
			if (Directory.Exists(Settings.ContentPath) == false)
			{
				try
				{
					// Create content directory if it doesn't exists the first time
					Directory.CreateDirectory(Settings.ContentPath);
				}
				catch (Exception ex)
				{
					// If that fails, this is a very bad error!
					Log.Warning("LoadContentList: Couldn't create the content " +
						"directory '" + Settings.ContentPath + "' because of:\n" + ex);
				}

				// obs.
				// -> now we fill it with data from the content server automatically
				//// We can ignore the "collecting request" because there are obviously
				//// no files in this newly created directory!
				//return;
			} // if
#elif !WINDOWS_PHONE
/*obs, not longer in settings!
			// On non WP7 platform (where file and directory access is forbidden),
			// we can at least check if the path is there and output a warning!
			if (Directory.Exists(Settings.ContentPath) == false)
			{
				Log.Warning("Content directory does not exist: " +
					Settings.Debug.ContentPath + ", unable to load Content.xml and " +
					"therefore all content loading will not work!");
				return;
			} // if
*/
#endif
			/*obs, not longer in settings!
#if DEBUG
			// Do a quick check if the Content directory is valid (we do not support
			// network share paths anymore).
			else if (Settings.Debug.ContentPath.StartsWith("\\"))
			{
				Log.Warning("The remote content path '" + Settings.Debug.ContentPath +
					"' from your settings is not longer supported, please use a local" +
					" path like e.g. 'Content'");
			}
#endif
*/
			// Initially we need to load the content list normally.
			LoadContentList(Settings.ContentPath);

#if LOG_STUFF
			Log.Info("ContentManager created: " + this);
#endif
		} // ContentManager()
		#endregion

		#region Destructor
		/// <summary>
		/// Destruct ContentManager, this is important if we don't call Dispose
		/// explicitly from Application.Close we still want the content to dispose
		/// in case there is new stuff we need to save to the Content.Data file!
		/// </summary>
		~ContentManager()
		{
			Dispose();
		} // ~ContentManager()
		#endregion

		#region CheckSettings
		/// <summary>
		/// This method checks if the settings haven't got a file yet.
		/// If this happens we check if we have a settings file in the content
		/// and then load and set the values to the settings.
		/// </summary>
		private void CheckSettings()
		{
			if (Settings.IsUncreated == false)
			{
				return;
			}

			// If we have a settings file in the content...
			if (contentDataDictionary.ContainsKey("Settings"))
			{
				// ...then load it and set it to the settings.
				XmlContent settingsContent = new XmlContent("Settings");
				Settings.SetContentDefaults(settingsContent.XmlData);
				settingsContent.Dispose();
			}
		} // CheckSettings()
		#endregion

//TODO: still needs to be fixed up and tested with Scene.Close!
		#region TryToKillUnusedContent
		/// <summary>
		/// Try to kill unused content
		/// </summary>
		private void TryToKillUnusedContent()
		{
			/*2011-03-09: disabled, causes too much problems with the scene and killing everything is not right anyway!
						// Kill all content classes, their internal native data is already gone!
						usedContentClasses.Clear();

						// Kill the native data and even the fallback data (which is also
						// linked in alreadyLoadedNativeData, no need to dispose it twice)
						foreach (NativeData data in alreadyLoadedNativeData.Values)
						{
							bool firstContentObject = true;
							object dataToDispose = null;
							foreach (Content content in data.usedContents)
							{
								// Make sure we clear all FailedToLoad flags (usually not set)
								content.data.FailedToLoad = false;
								// And dispose all the native content (usually only kills the data
								// once and then will just null out caches for all other instances)
								if (firstContentObject)
								{
									firstContentObject = false;
									dataToDispose = content.CachedData;
								}
								//don't do this here, causes problems, behaves differently on each
								//platform: content.DisposeData(dataToDispose);
								dataToDispose = null;
							} // foreach
						} // foreach
						alreadyLoadedNativeData.Clear();
						cachedFallbackData.Clear();

						// And all the content data is gone too
						contentDataDictionary.Clear();
			 */
		} // TryToKillUnusedContent()
		#endregion

//TODO: still needs to be fixed up!
		#region Dispose (virtual)
		/// <summary>
		/// Dispose the content manager and all content. Any future request to
		/// get ContentData from the ContentList will result in warnings and the
		/// creation of new ContentData objects without using the real content.
		/// This is only called at the end of the application.
		/// </summary>
		public virtual void Dispose()
		{
			// Kill all content classes, their internal native data is already gone!
			usedContentClasses.Clear();

			// Kill the native data and even the fallback data (which is also
			// linked in alreadyLoadedNativeData, no need to dispose it twice)
			foreach (NativeData data in alreadyLoadedNativeData.Values)
			{
				bool firstContentObject = true;
				object dataToDispose = null;
				foreach (Content content in data.usedContents)
				{
					// Make sure we clear all FailedToLoad flags (usually not set)
					content.data.FailedToLoad = false;
					// And dispose all the native content (usually only kills the data
					// once and then will just null out caches for all other instances)
					if (firstContentObject)
					{
						firstContentObject = false;
						dataToDispose = content.CachedData;
					}
					content.DisposeInternalData(dataToDispose);
					dataToDispose = null;
				} // foreach
				data.data = null;
			} // foreach
			alreadyLoadedNativeData.Clear();
			cachedFallbackData.Clear();

			GC.Collect();
			GC.WaitForPendingFinalizers();
		} // Dispose()
		#endregion

		#region LoadContentList (abstract)
		/// <summary>
		/// Load ContentList, called from the constructor and every time the
		/// content list needs to be updated (e.g. new content was detected in
		/// RawContent and copied to the PlatformContent and then locally here).
		/// Implemented in the concrete ContentManager class to allow better
		/// updates and content refreshes.
		/// </summary>
		protected abstract void LoadContentList(string localContentPath);
		#endregion

		#region OnLanguageChanged (virtual)
		/// <summary>
		/// On language changed, called whenever the language has been changed.
		/// This forces all content that has the IsLanguageDependant flag set to
		/// search for a new language fitting content data and relinking in each
		/// Content class that uses it.
		/// </summary>
		protected virtual void OnLanguageChanged()
		{
#if LOG_STUFF
			Log.Info("ContentManager.OnLanguageChanged was called");
#endif
			// Go through all content that is already initialized (this does not
			// mean loaded, just that the Content constructor called GetContentData)
			// Just going through all usedContentClasses works too, but is way slower
			// and might introduce new issues when changing the language (reloading
			// files way too often because they are used multiple times). Instead we
			// just collect all the content data that needs to be changed and then
			// just update those (and the used content classes .data)
			Dictionary<ContentData, List<Content>> contentThatNeedsUpdating =
				new Dictionary<ContentData, List<Content>>();
			foreach (Content content in usedContentClasses)
			{
				// We only need to update language dependant content data
				if (content.data.IsLanguageDependant)
				{
					// Check if we already have checked this content data
					List<Content> contents = null;
					if (contentThatNeedsUpdating.TryGetValue(content.data,
						out contents) == false)
					{
						// No? Then create new list
						contents = new List<Content>();
						contentThatNeedsUpdating.Add(content.data, contents);
					}
					// In any case, just add this class to the list
					contents.Add(content);

					// Dispose the old language content to free up some memory
					// Note: This only has to be done once per content data!

					//cannot be deleted because this is most likely an atlas image and
					//then we kill other images in that atlas too and we are skrewed:
					//content.DisposeData(content.CachedData);
					//BUUUUUUUUUT if we had a different than english language, we
					// should be safe to kill all native data because we only want
					// english data from now on
					/*also does not work, grr
					if (Settings.Language == Localization.English)
					{
						content.DisposeData(content.CachedData);
					}
					 */

					content.ForceReload();
				} // if
			} // foreach

			//TODO: re-enable: Rest is disabled for now, only this way language changes work currently!
#if OBS
			return;
			// Okay, it is GC.Collect time one again (we might just have cleared
			// out some content objects)
			GC.Collect();


			// Now that we have everything collected, update all content data
			// (which should be much less than the usedContentClasses that had
			// language dependant content in the first place, but we update all
			// those classes .data as well once we reloaded it).
			foreach (ContentData data in contentThatNeedsUpdating.Keys)
			{
#if LOG_STUFF
				Log.Info("OnLanguageChanged: Updating old content data " + data);
#endif
				// Only grab newData once (does not matter how many classes use it)
				ContentData newData = null;
				List<Content> contents = contentThatNeedsUpdating[data];
				foreach (Content content in contents)
				{
#if LOG_STUFF
					Log.Info("OnLanguageChanged: Updating old content class " + content);
#endif
					// Link up to the new localized language content (should exist)!
					if (newData == null)
					{
						newData = GetContentData(content, content.data.Name,
							content.data.Type);
					}
					// And finally link up the newData with each content class that
					// uses it.
					content.data = newData;
//TODO: this could be a major bug, we need to force a new load if the old data was already loaded? still true? yes
				} // foreach
			} // foreach
#endif
		} // OnLanguageChanged()
		#endregion

		#region CheckIfContentIsStillUpToDate (abstract)
		/// <summary>
		/// If you want to check if all local content files are still up to date
		/// (comparing them with the RawContent files) use this feature. It has
		/// some performance implications, so you can disable it in the Settings
		/// (default is disabled for Release builds)!
		/// </summary>
		protected virtual ContentData CheckIfContentIsStillUpToDate(
			ContentData data)
		{
			// By default just return data again.
			return data;
		} // CheckIfContentIsStillUpToDate(data)
		#endregion

		#region GetContentData (virtual)
		/// <summary>
		/// Search for content data with a given name and type (that is all we
		/// know when trying to initialize content). This is used to load the
		/// actual content via the LocalFilePath property, but we can also use
		/// lots of other meta-data that is provided through ContentData even
		/// before actually loading the file via Content.Load.
		/// </summary>
		protected internal virtual ContentData GetContentData(Content contentClass,
			string name, ContentType type)
		{
			#region Empty content name check
			// We can skip all this stuff if the name starts with the '<' character,
			// e.g. "<IntroScene>" is a code generated scene and has no content!
			// If the name was empty (see Content.DefaultId), also skip all the
			// checking and return a dummy ContentData object too.
			if (String.IsNullOrEmpty(name) ||
				name == Content.DefaultId ||
				name.StartsWith("<"))
			{
				/*this is probably not required, we never need to update this!
				// Since we don't call base.GetContentData, check ourselves if this
				// class is already in the usedContentClasses list for updates.
				if (usedContentClasses.Contains(contentClass) == false)
				{
					usedContentClasses.Add(contentClass);
				}
				 */

				// Make sure we assign a real string and not null, name is used often!
				if (name == null)
				{
					name = Content.DefaultId;
				}

				// Just generate dummy data, we won't cache this, future request
				// each get their own dummy data, but we will never update this
				// because there is no real content behind this content request!
				ContentData dummyData = new ContentData(name, type, English,
					// Note: The LocalFilePath is empty, only LoadFallbackData is used.
					"", projectName, false, 0, DateTime.Now);
				if (type == ContentType.Image)
				{
					// Set information about the fallback image already here so we
					// do not get confusing warnings when the meta data is set later.
					dummyData.SetImageMetaData(new Size(4, 4), BlendMode.Opaque, false,
						Rectangle.One, Rectangle.One, 0, 0, null);
				}
				return dummyData;
			} // if
			#endregion

			#region Add to usedContentClasses
			// Add the content class to the usedContentClasses list, this is used
			// for language changes and content updates to re-link the Data field.
			// Note: This is also called for OnLanguageChanged, which just updates
			// the data field, no reason to add it again!

			// NOTE: Remove it for soulcraft, we use our own localization.
			//if (usedContentClasses.Contains(contentClass) == false)
			//{
			//  usedContentClasses.Add(contentClass);
			//}
			#endregion

			#region Try to find in local content
			// Try to find content based on the name, type and current language.
			// Note: This might look more complicated than just going through a
			// simple flat list, but this is 10x times faster for a few hundred
			// content files and even better for more content files (100-1000x).
			// While this might seem not so important since it is only important for
			// the startup time when loading content, if you think about the number
			// of string compares required when using flat lists, it can get bad
			// very fast, just 1000 files (e.g. ZombieParty) and then comparing
			// those 1000 content loads with the 1000 raw content files is already
			// 1 million string compares, that is just too much. With dictionaries
			// this goes down to just a few hash checks and is much faster!
			List<ContentData> datas = null;
			if (contentDataDictionary.TryGetValue(name, out datas))
			{
				foreach (ContentData data in datas)
				{
					if (data.Type == type &&
						data.Language == Settings.Language)
					{
						// Return data, but check if file is not up to date, then renew!
						return CheckIfContentIsStillUpToDate(data);
					}
				} // foreach

				// If content data was not found yet, use the English language content!
				if (Settings.Language != English)
				{
					foreach (ContentData data in datas)
					{
						if (data.Type == type &&
							data.Language == English)
						{
							// Return data, but check if file is not up to date, then renew!
							return CheckIfContentIsStillUpToDate(data);
						}
					} // foreach
				} // if

			} // foreach

#if LOG_STUFF
			Log.Info("GetContentData: No local data for name '" + name +
				"' and type '" + type + "' for content class found " +
				contentClass);
#endif
			#endregion

			// Fallback code must be implemented in derived class, null is not valid
			return null;
		} // GetContentData(contentClass, name, type)
		#endregion

		#region GetAllContentDatasForAType
		/// <summary>
		/// Get all content datas for a type helper method. Used to find out which
		/// content files are available for one type. Used only internally in
		/// the ContentSystem. Currently used for shaders to find out which
		/// shaders are available and to allow loading them by name from the
		/// ShaderFeatureFlags constructor, which calls this and returns the
		/// ContentDatas were it can search for a shader with this features. If
		/// nothing was found it still can provide a fallback.
		/// </summary>
		public static List<ContentData> GetAllContentDatasForAType(
			ContentType contentType)
		{
			// Usually the content list does not change, we can cache this.
			return Cache<ContentType, List<ContentData>>.GetCachedData(contentType,
				delegate(ContentType inputContentType)
				{
					List<ContentData> ret = new List<ContentData>();

					foreach (List<ContentData> datas in
						Instance.contentDataDictionary.Values)
					{
						foreach (ContentData data in datas)
						{
							if (data.Type == inputContentType)
							{
								ret.Add(data);
							}
						}
					} // foreach

					return ret;
				});
		} // GetAllContentDatasForAType(contentType)
		#endregion

		#region GetImageAnimationFrameNames
		/// <summary>
		/// Get all frame names of an image. Returns nothing if the image name
		/// was not found, 1 frame for no animation, just an image and for more
		/// frames it is an animated image.
		/// </summary>
		public static string[] GetImageAnimationFrameNames(string imageName)
		{
			// First get the initial content data
			List<ContentData> allImages =
				GetAllContentDatasForAType(ContentType.Image);
			foreach (ContentData imageContent in allImages)
			{
				if (StringHelper.Compare(imageName, imageContent.Name))
				{
					// We found the image directly? Then just return it!
					return new string[] { imageContent.Name };
				}
			}

			// Otherwise grab all images that start with this name and end with
			// a number!
			List<string> imageNames = new List<string>();
			foreach (ContentData imageContent in allImages)
			{
				// Is this the name we are searching for?
				if (StringHelper.StartsWith(imageContent.Name, imageName) &&
					// Okay, then cut off the number and make sure the name still fits
					StringHelper.Compare(imageName,
					StringHelper.RemoveNumbersAtEnd(imageContent.Name)))
				{
					// Okay, just add it to our imageNames list (will be already sorted
					// by the content generation process).
					imageNames.Add(imageContent.Name);
				} // if
			} // foreach

			if (imageNames.Count == 0)
			{
				Log.Warning("Image or animation not found: " + imageName);
			}

			// Return whatever we found, can be nothing (empty list) or all the
			// animated images.
			return imageNames.ToArray();
		} // GetImageAnimationFrameNames()
		#endregion

		#region RemoveContentClass
		/// <summary>
		/// Remove content data, this makes sure after disposing content we also
		/// kill all the links to the ContentManager and we make sure the Data is
		/// not longer used (if it is not used by other content instances anymore).
		/// </summary>
		internal void RemoveContentClass(Content contentClass)
		{
			try
			{
				bool rememberToRemoveEmptyContentClasses = false;
				foreach (NativeData data in alreadyLoadedNativeData.Values)
				{
					bool found = false;
					foreach (Content content in data.usedContents)
					{
						if (content == contentClass)
						{
							// Found content usage, try to remove it
							found = true;
							break;
						} // if
					} // foreach
					if (found)
					{
						data.usedContents.Remove(contentClass);

						// Do we have zero content classes remaining for this native
						// data?
						if (data.usedContents.Count == 0)
						{
							// Then we should fully dispose it, it is not longer needed!
							contentClass.DisposeInternalData(data.data);
							// Also remove this from alreadyLoadedNativeData.
							rememberToRemoveEmptyContentClasses = true;
						}
					}
				} // foreach

				// This is a little complicated, but we still need to remove the now
				// empty list of used content classes if we just caused such a case.
				if (rememberToRemoveEmptyContentClasses)
				{
					foreach (string key in alreadyLoadedNativeData.Keys)
					{
						if (alreadyLoadedNativeData[key].usedContents.Count == 0)
						{
							alreadyLoadedNativeData.Remove(key);
							break;
						}
					}
				} // if

				// In any case, remove this from the used content classes.
				usedContentClasses.Remove(contentClass);
			} // try
			catch
			{
#if DEBUG
				//Log.Warning("TODO: Currently crashes on iPad when disposing Music");
#endif
			}
		} // RemoveContentClass(contentClass)
		#endregion

		#region Run (override)
		/// <summary>
		/// Run method, called every tick by the Application Run method. Does not
		/// only handle streaming for all content, but also will load all the
		/// content that has not been loaded yet (if flagged to be loaded as soon
		/// as possible in the constructor or if triggered through the initialize
		/// method here).
		/// </summary>
		public override void Run()
		{
			CheckSettings();

			//TODO: threading should happen here (also see Content constructor), use extra loading thread that usually sleeps

			// Check all content if we want to load anything.
			foreach (Content content in usedContentClasses)
			{
				//TODO: this can be optimized, loadInRun is usually only set in the first frame, then we do not need to check every frame!
				// We only care about content that still has loadInRun set!
				if (content.loadInRun)
				{
					content.ForceLoad();
				}
			} // foreach

			//TODO: we should do streaming here too!
		} // Run()
		#endregion
	} // class ContentManager
} // namespace Delta.ContentSystem
#if OBS
//TODO: put this into the MonoTouch CsprojConverter part, just convert all Dictionaries that cannot be used with aot to ones that use a special comparer!
	//tst: proving an IEqualityComparer for contentDataDictionary
	// This is to avoid the iPhone crashes because MonoTouch is jit'ing stuff:
/*Wed Oct  6 05:28:46 iPad-exDream UIKitApplication:de.mobilebits.ZombieParty[0xf994][924] <Notice>: 00.022 Warning: Fatal Error: Unhandled exception occurred (Application is terminated, we cannot recover!): System.TypeInitializationException: An exception was thrown by the type initializer for ZombieParty.Players.Player ---> System.Reflection.TargetInvocationException: Exception has been thrown by the target of an invocation. ---> System.ExecutionEngineException: Attempting to JIT compile method 'Delta.ContentSystem.Client.ClientContentManager:GetContentData (Delta.ContentSystem.Content,string,Delta.ContentSystem.ContentType)' while running with --aot-only.
	*/
/*not longer required for MonoTouch 3.2
	public class ContentNameComparer : IEqualityComparer<string>
	{
#region IEqualityComparer<string> Members
		public bool Equals(string x, string y)
		{
			return String.Compare(x, y, StringComparison.CurrentCultureIgnoreCase) == 0;
		}

		public int GetHashCode(string obj)
		{
			return obj.GetHashCode();
		}
#endregion
	}

	public class ContentTypeComparer : IEqualityComparer<ContentType>
	{
#region IEqualityComparer<string> Members
		public bool Equals(ContentType x, ContentType y)
		{
			return x == y;
		}

		public int GetHashCode(ContentType obj)
		{
			return obj.GetHashCode();
		}
#endregion
	}
 */
	/*not longer required
			/// <summary>
			/// TOCO
			/// </summary>
			private static ContentNameComparer nameComparer =
				new ContentNameComparer();
			/// <summary>
			/// TOCO
			/// </summary>
			private static ContentTypeComparer typeComparer =
				new ContentTypeComparer();
	*/
//TODO: for profiler
		/*
#region LoadContent
		/// <summary>
		/// Loads the native content data.
		/// </summary>
		/// <returns>The native content object.</returns>
		private object LoadContent(string relativeContentPath)
		{
			// Disabled in release mode or if not in windows (this is very slow)!
#if DEBUG && WINDOWS
			// And also only do this if we have the System profiling mode enabled,
			// which shows the content data loaded on the screen.
			if (relativeContentPath != Content.UnknownPath &&
				Settings.Debug.IsProfilingModeOn(DebugMode.System))
			{
				// Quick and very simple logic to determine how "big" the memory size
				// is that every content object needs (maybe there is a lot better
				// way?) -> can be used for (memory) profiling
				// If the current content object is a default object
				int DataSize = (Name == DefaultContentName ||
					File.Exists(relativeContentPath) == false) ?
					// we currently assume 100 KB as content size for it
					// Note: This is also used for atlases when files cannot be found!
					100 * 1024 :
					// else we just use the file size (for the moment)
					// Note: Now only happens if profiling is on, which is only used
					// on Windows because FileInfo is way too slow for many files.
					(int)new FileInfo(relativeContentPath).Length;

				//TODO: check for more efficient ways to find profiling module
				IProfiler profiler = (IProfiler)Factory.Create(typeof(IProfiler));
				if (profiler != null)
				{
					profiler.AddLoadedContentSize(DataSize);
				}
			} // if
#endif
		 */
#endregion
#region CacheSize
		/// <summary>
		/// Returns the size of the content cache
		/// </summary>
		int CacheSize { get; } // CacheSize
#endregion

//TODO: rename!
#region GetPath
		/// <summary>
		/// Gets the (relative) path where the content file is located.
		/// </summary>
		string GetPath(string contentName, ContentType type);
#endregion

#region DoesContentFileExist
		/// <summary>
		/// Returns 'true' if the file exists in the content (folder).
		/// </summary>
		/// <param name="relativePath">Relative path based on the content folder.
		/// </param>
		/// <returns></returns>
		bool DoesContentFileExist(string relativePath, ContentType type);
#endregion

//TODO: remove this again, currently only public for Font generation,
// which needs to add more content while the app is running.
#region AddContentFileInfo
		/// <summary>
		/// Add content file info to the overall contentFilePaths dictionary.
		/// </summary>
		void AddContentFileInfo(string newPath, ContentType type);
#endregion

//TODO: describe why this is needed, we need this if the same data is loaded twice!
#region GetLoadedData
		/// <summary>
		/// Gets the already loaded content file data (-> getting content cache).
		/// </summary>
		/// <param name="contentName">Content name</param>
		/// <param name="contentType">Content type</param>
		/// <returns>Object</returns>
		object GetLoadedData(string contentName, ContentType contentType);
#endregion

#region SetLoadedData
		/// <summary>
		/// Sets the loaded content file data (-> updating content cache).
		/// </summary>
		/// <param name="contentName">Content name</param>
		/// <param name="contentType">Content type</param>
		/// <param name="newLoadedData">New loaded data</param>
		void SetLoadedData(string contentName, ContentType contentType,
			object newLoadedData);
#endregion

		//Helpers for Atlas functionality, TODO: rethink
		bool IsTextureAtlasValid(string name);

		Rectangle GetAtlasTextureUV(string name, out BlendMode blendMode);

// Expose ContentManager internals to Delta.ContentSystem.Client
//[assembly: System.Runtime.CompilerServices.InternalsVisibleTo(
//	"Delta.ContentSystem.Client")]

//TODO: use localizationManager from Application.Localization.
/*TODO: reimplement
#region LocaStringManager
//obs: ILoca LocaManager { get; }
#endregion
 */

#region Save
		/// <summary>
		/// Creates the file at the given path and return the handle for saving the
		/// data.
		/// Note:
		/// Please make sure that the handle will be closed and released again after
		/// finishing storing.
		/// </summary>
		/// <param name="relativePath">Relative path to the content folder.</param>
		StreamWriter Save(string relativePath);
#endregion

#region GetAllContentFilesOf
		/// <summary>
		/// Get all content files of
		/// </summary>
		string[] GetAllContentFilesOf(ContentType type);
#endregion

		//TODO: IModule, then also use ApplicationModule as parent (see DynamicBase.Content) and have some Run code! all content could be updated in the Run method here?
/*obs
#region RootPath
		/// <summary>
		/// The root path of the content
		/// </summary>
		string RootPath { get; } // RootPath
#endregion
*/
		/*TODO: remove, obs, not really used (was for streaming, but that should work different anyway)!
#region RegisterAsUser
		/// <summary>
		/// Registers a Content object as user of a certain content in the
		/// contentCache.
		/// </summary>
		/// <param name="type">The type of the content being registered.</param>
		/// <param name="name">The name of the content being registered.</param>
		/// <param name="user">The user of the content.</param>
		void RegisterAsUser(ContentType type, string name, IBaseContent user);
#endregion

#region UnregisterAsUser
		/// <summary>
		/// Unregisters a Content object as user of a certain content in the
		/// contentCache.
		/// </summary>
		/// <param name="type">The type of the content being unregistered.</param>
		/// <param name="name">The name of the content being unregistered.</param>
		/// <param name="user">The user of the content.</param>
		void UnregisterAsUser(ContentType type, string name, IBaseContent user);
#endregion

#region GetNumberOfRegisteredUsers
		/// <summary>
		/// Gets the number of users currently assigned to a specific content.
		/// </summary>
		/// <param name="type">Type</param>
		/// <param name="name">Name</param>
		int GetNumberOfRegisteredUsers(ContentType type, string name);
#endregion

#region ClearUnusedData
		/// <summary>
		/// Calling this will let the ContentManager clear all data that no longer
		/// have registered users.
		/// </summary>
		void ClearUnusedData();
#endregion

#region Obs. stuff (can be reactived later...?)
		//TODO: add IModule loading stuff here (remove from Base)!

		//void Disconnect();

		///// <summary>
		///// Helper for ContentLoader to figure out where the content file is
		///// now located on disc. If loading is not possible right away an empty
		///// string is returned and we have to wait for the FileAvailable event!
		///// </summary>
		///// <param name="contentName">
		///// The content name in the format [File] or [Project]/[File],
		///// e.g. "Background" or "GravitySwipe/Background"
		///// </param>
		///// <param name="FileAvailable">
		///// FileAvailable delegate, which if used (will never be called if
		///// GetLocalFilename already returned a valid local filename) returns the
		///// local filename once the loading is complete.
		///// </param>
		///// <returns>
		///// The local filename we can use for loading this file or an empty string
		///// if loading is not yet possible (no content or it is still downloading)
		///// </returns>
		//void Load(string contentName, ContentType type,
		//  FileReloadDelegate fileLoadDelegate);

		///// <summary>
		///// Prepare loading in the content manager, this way the content can
		///// be downloaded while the app is starting. Later when the content
		///// is used for the first time, the actual loading happens and depending
		///// if the file is available yet (usually it is because it was loaded
		///// last session) the file or the dummy content will be loaded until
		///// the FileAvailableDelegate above returns the new filename.
		///// </summary>
		///// <param name="contentName">
		///// The content name in the format [File] or [Project]/[File],
		///// e.g. "Background" or "GravitySwipe/Background"
		///// </param>
		///// <param name="type">Type of content (image, sound, etc.)</param>
		//void PrepareLoad(string contentName, ContentType type);

		/// <summary>
		/// Get local file name
		/// </summary>
		/// <param name="contentName">the content name [Project]/[Contentname] or
		///  [Contentname] eg. "ProjectXY/Tree" or "Tree"</param>
		///  <param name="type">the content type eg. Image</param>
		//obs:	string GetLocalFileName(string contentName, ContentType type);

		// obs
		///// <summary>
		///// Since the content manager keeps a cache list of all the loaded content
		///// it is the easiest way to reload all the content in case we changed
		///// some module, lost device focus or just for fun to refresh all data.
		///// Please note that each content class can be used to reload specific
		///// content anyway, we can reload images, shaders, etc. on demand. Also
		///// if the content file changes on the content server, we will get
		///// notified and the content will be reloaded in the same matter. This
		///// way if an artist overwrite any content file, it will automatically be
		///// reloaded in all running apps using that content file!
		///// </summary>
		//void ReloadAllContent();
#endregion

#endif
