using System;

namespace ProconLDS.Configuration
{
    /// <summary>
    /// AlarmControl status.
    /// </summary>
    [Flags]
    public enum AlarmControlStatus
    {
        /// <summary>
        /// Unknown status.
        /// </summary>
        Unknown = 0,

        /// <summary>
        /// Alarm is cleared.
        /// </summary>
        Cleared = 1,

        /// <summary>
        /// Alarm is active.
        /// </summary>
        Active = 2,

        /// <summary>
        /// Alarm is unacknowledged.
        /// </summary>
        Unacknowledged = 4,

        /// <summary>
        /// Alarm is acknowledged.
        /// </summary>
        Acknowledged = 8,

        /// <summary>
        /// Alarm is active and unacknowledged.
        /// </summary>
        ActiveUnack = Active + Unacknowledged,

        /// <summary>
        /// Alarm is active but acknowledged.
        /// </summary>
        ActiveAck = Active + Acknowledged,

        /// <summary>
        /// Alarm is cleared but unacknowledged.
        /// </summary>
        ClearedUnack = Cleared + Unacknowledged,
    }

    /// <summary>
    /// Alarm status management class.
    /// </summary>
    public class AlarmControl
    {
        #region Fields

        private bool _alarm;

        #endregion Fields

        #region Events

        /// <summary>
        /// Raised when the alarm is acknowledged.
        /// </summary>
        public event EventHandler AlarmAck;

        /// <summary>
        /// Raised when the alarm is cleared.
        /// </summary>
        public event EventHandler AlarmClear;

        /// <summary>
        /// Raised when the alarm is set.
        /// </summary>
        public event EventHandler AlarmSet;

        #endregion Events

        #region Properties

        /// <summary>
        /// True if active alarm acknowledged.
        /// </summary>
        public bool Acknowledged { get; private set; }

        /// <summary>
        /// Get/Set the current alarm status.
        /// </summary>
        public bool Alarm
        {
            get { return _alarm; }
            set
            {
                // only do anything if value has changed
                if (_alarm != value && !Inhibit)
                {
                    // record time of event
                    LastEvent = DateTime.Now;
                    // set new alarm value
                    _alarm = value;
                    // set unacknowledged if alarm set
                    if (_alarm)
                    {
                        UnAcknowledged = true;
                        // signal alarm set to others
                        AlarmSet?.Invoke(this, EventArgs.Empty);
                    }
                    // clear acknowledged if alarm cleared
                    else
                    {
                        Acknowledged = false;
                        // signal alarm cleared to others
                        AlarmClear?.Invoke(this, EventArgs.Empty);
                    }
                }
            }
        }

        /// <summary>
        /// Message displayed when this alarm is active and/or unacknowledged.
        /// </summary>
        public string AlarmMessage { get; set; } = string.Empty;

        /// <summary>
        /// True to inhibit activation of this alarm
        /// </summary>
        public bool Inhibit { get; set; }

        /// <summary>
        /// Date/time of last event.
        /// </summary>
        public DateTime LastEvent { get; set; } = DateTime.Now;

        /// <summary>
        /// Current status of the alarm.
        /// </summary>
        public AlarmControlStatus Status
        {
            get
            {
                var status = Alarm ? AlarmControlStatus.Active : AlarmControlStatus.Cleared;
                if (UnAcknowledged)
                    status |= AlarmControlStatus.Unacknowledged;
                if (Acknowledged)
                    status |= AlarmControlStatus.Acknowledged;
                return status;
            }
        }

        /// <summary>
        /// True if alarm not acknowledged.
        /// </summary>
        public bool UnAcknowledged { get; private set; }

        #endregion Properties

        #region Methods

        /// <summary>
        /// Acknowledges an active alarm.
        /// </summary>
        public void AckAlarm()
        {
            var doAckEvent = UnAcknowledged;
            // clear unacknowledged
            UnAcknowledged = false;
            // if still active set acknowledged
            if (_alarm)
                Acknowledged = true;
            // record time of event
            LastEvent = DateTime.Now;
            // raise acknowledged event if required
            if (doAckEvent)
            {
                AlarmAck?.Invoke(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// String representation of this alarm.
        /// </summary>
        public override string ToString()
        {
            var status = Alarm ? "Active" : "Cleared";
            if (UnAcknowledged)
                status += " Unacknowledged";
            if (Acknowledged)
                status += " Acknowledged";
            return $"{LastEvent:yyyy-MM-dd HH:mm:ss,fff} : {AlarmMessage} {status}";
        }

        #endregion Methods
    }
}